from .http_config import HTTPConfig
from youtube_transcript_api.proxies import ProxyConfig, GenericProxyConfig, WebshareProxyConfig
from .logging_config import enable_default_config

__all__ = [
    "HTTPConfig",
    "enable_default_config",
    "ProxyConfig",
    "GenericProxyConfig",
    "WebshareProxyConfig"
]


from fake_useragent import UserAgent
import random

ACCEPT_LANGUAGES = [
    "en-US,en;q=0.9",
    "en-GB,en;q=0.9",
    "fr-FR,fr;q=0.9"
]

ACCEPT = [
    "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8",
    "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8",
    "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
    "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9"
]

REFERERS = [
    "https://www.youtube.com/",
    "https://www.google.com/",
    "https://www.bing.com/",
    "https://duckduckgo.com/"
]

SEC_FETCH_DEST = ["document", "empty"]
SEC_FETCH_MODE = ["navigate", "no-cors"]
SEC_FETCH_SITE = ["none", "same-origin", "cross-site"]

def get_realistic_headers() -> dict:
    """
    Creates realistic headers for mimic browser behavior which reduces the changes of getting banned immediatly.\n
    Uses `fake_useragent` package for creating random user agents.
    """
    ua = UserAgent(platforms='desktop', os='Windows')
    user_agent = ua.random
    return {
        "User-Agent": user_agent,
        "Accept": random.choice(ACCEPT),
        "Accept-Language": random.choice(ACCEPT_LANGUAGES),
        "Referer": random.choice(REFERERS),
        "Connection": "keep-alive",
        "DNT": "1",
        "Upgrade-Insecure-Requests": "1",
        "Accept-Encoding": "gzip, deflate",

        "Sec-Fetch-Dest": random.choice(SEC_FETCH_DEST),
        "Sec-Fetch-Mode": random.choice(SEC_FETCH_MODE),
        "Sec-Fetch-Site": random.choice(SEC_FETCH_SITE),
        "Sec-CH-UA-Platform": "\"Windows\"",
        "Sec-CH-UA": f"\"{random.choice(['Chromium', 'Google Chrome', 'Not.A/Brand'])}\";v=\"{random.randint(100, 135)}\"",
    }



import httpx
from ytfetcher.utils.headers import get_realistic_headers
from ytfetcher.exceptions import InvalidHeaders

class HTTPConfig:
    """
    Configuration object for HTTP client settings.

    This class provides a structured way to configure HTTP-related options 
    such as timeouts and headers when making network requests. It ensures 
    that headers are valid and assigns default, realistic browser-like headers 
    if none are provided.

    Attributes:
        timeout (httpx.Timeout): 
            Timeout configuration for HTTP requests.
        
        headers (dict): 
            Dictionary of HTTP headers to be used in requests.
    """
    def __init__(self, timeout: float | None = None, headers: dict | None = None):
        self.timeout = httpx.Timeout(timeout=timeout) or httpx.Timeout()
        self.headers = headers or get_realistic_headers()

        if not isinstance(headers, dict | None):
            raise InvalidHeaders("Invalid headers.")


class YTFetcherError(Exception):
    """
    Base exception for all YTFetcher errors.
    """

class ExporterError(Exception):
    """
    Base exception for all Exporter errors.
    """

class OutputDirectoryNotFoundError(ExporterError):
    """
    Raised when the specified output directory does not exist.
    """

class NoDataToExport(ExporterError):
    """
    Raises when channel snippets and transcripts are empty.
    """

class InvalidHeaders(YTFetcherError):
    """
    Raises when headers are invalid.
    """


from ytfetcher._youtube_dl import get_fetcher
from ytfetcher.models.channel import ChannelData, DLSnippet
from ytfetcher._transcript_fetcher import TranscriptFetcher
from ytfetcher._youtube_dl import CommentFetcher
from ytfetcher.config.http_config import HTTPConfig
from youtube_transcript_api.proxies import ProxyConfig
from typing import Iterable

class YTFetcher:
    """
    YTFetcher is a high-level interface for fetching YouTube video metadata and transcripts.

    It supports three modes of initialization:
    - From a channel handle (via `from_channel`)
    - From a playlist ID (via `from_playlist_id`)
    - From a list of specific video IDs (via `from_video_ids`)

    Internally, it uses the yt-dlp to retrieve video snippets and metadata,
    and the `youtube_transcript_api` (with optional proxy support) to fetch transcripts.

    Args:
        max_results (int): Maximum number of videos to fetch.
        video_ids (list[str]): List of specific video IDs to fetch.
        playlist_id (str | None) Playlist id to fetch from.
        channel_handle (str | None): Optional YouTube channel handle (used when fetching from channel).
        http_config (HTTPConfig): Configuration for HTTP client behavior.
        proxy_config (ProxyConfig | None): Optional proxy settings for transcript fetching.
        languages (Iterable[str]): Preferred languages to fetch first, default to `en`
        manually_created (bool): Flag for fetching only manually created transcripts. Default to False
    """
    def __init__(
        self,
        max_results: int,
        video_ids: list[str] | None,
        playlist_id: str | None = None,
        channel_handle: str | None = None,
        proxy_config: ProxyConfig | None = None,
        http_config: HTTPConfig = HTTPConfig(),
        languages: Iterable[str] = ("en", ),
        manually_created: bool = False,
        ):
        self.http_config = http_config
        self.proxy_config = proxy_config

        self.youtube_dl = get_fetcher(channel_handle, playlist_id, video_ids, max_results)
        self.snippets = self.youtube_dl.fetch()

        self.fetcher = TranscriptFetcher(self._get_video_ids(), http_config=self.http_config, proxy_config=self.proxy_config, languages=languages, manually_created=manually_created)
    
    @classmethod
    def from_channel(
        cls,
        channel_handle: str,
        max_results: int = 50,
        http_config: HTTPConfig = HTTPConfig(),
        proxy_config: ProxyConfig | None = None,
        languages: Iterable[str] = ("en",),
        manually_created: bool = False
        ) -> "YTFetcher":
        """
        Create a fetcher that pulls up to max_results from the channel.
        """
        return cls(
            http_config=http_config,
            max_results=max_results,
            video_ids=None,
            channel_handle=channel_handle,
            proxy_config=proxy_config,
            languages=languages,
            manually_created=manually_created
            )
    
    @classmethod
    def from_video_ids(
        cls,
        video_ids: list[str] = [],
        http_config: HTTPConfig = HTTPConfig(),
        proxy_config: ProxyConfig | None = None,
        languages: Iterable[str] = ("en",),
        manually_created: bool = False
        ) -> "YTFetcher":
        """
        Create a fetcher that only fetches from given video ids.
        """
        return cls(
            http_config=http_config,
            max_results=len(video_ids),
            video_ids=video_ids,
            channel_handle=None,
            proxy_config=proxy_config,
            languages=languages,
            manually_created=manually_created
            )
    
    @classmethod
    def from_playlist_id(
        cls,playlist_id: str,
        max_results: int = 50,
        http_config: HTTPConfig = HTTPConfig(),
        proxy_config: ProxyConfig | None = None,
        languages: Iterable[str] = ("en",),
        manually_created: bool = False
        ) -> "YTFetcher":
        """
        Create a fetcher tthat fetches from given playlist id.
        """
        return cls(
            http_config=http_config,
            playlist_id=playlist_id,
            proxy_config=proxy_config,
            languages=languages,
            max_results=max_results,
            video_ids=None,
            manually_created=manually_created
            )

    async def fetch_youtube_data(self) -> list[ChannelData]:
        """
        Asynchronously fetches transcript and metadata for all videos retrieved from the channel or video IDs.

        Returns:
            list[ChannelData]: A list of objects containing transcript text and associated metadata.
        """

        transcripts = await self.fetcher.fetch()
        
        for transcript, snippet in zip(transcripts, self.snippets):
            transcript.metadata = snippet if transcript.transcripts else None
        
        return transcripts
    
    async def fetch_with_comments(self, max_comments: int = 20, max_workers: int = 30) -> list[ChannelData]:
        """
        Fetches comments, addition to transcripts and metadata.

        Args:
            max_comments: Max number of comments to fetch.
            max_workers: Max number of workers for threads.

        Returns:
            list[ChannelData]: A list objects containing transcript text, metadata and comments.
        """

        transcripts = await self.fetcher.fetch()
        
        commf = CommentFetcher(max_comments=max_comments, max_workers=max_workers)
        full_comments = commf.fetch(video_ids=self._get_video_ids())

        for transcript, snippet, comments in zip(transcripts, self.snippets, full_comments):
            transcript.metadata = snippet if transcript.transcripts else None
            transcript.comments = comments
        
        return transcripts
    
    async def fetch_comments(self, max_comments: int = 20, max_workers: int = 30) -> list[ChannelData]:
        """
        Fetches comments for all videos.

        Args:
            max_comments: Max number of comments to fetch.
            max_workers: Max number of workers for threads.

        Returns:
            list[ChannelData]: A list of objects containing only comments.
        """
        commf = CommentFetcher(max_comments=max_comments, max_workers=max_workers)
        full_comments = commf.fetch(video_ids=self._get_video_ids())

        return [
            ChannelData(
                video_id=snippet.video_id,
                transcripts=None,
                metadata=snippet,
                comments=comments
            )
            for snippet, comments in zip(self.snippets, full_comments)
        ]
    
    async def fetch_transcripts(self) -> list[ChannelData]:
        """
        Returns only the transcripts from cached or freshly fetched YouTube data.

        Returns:
            list[ChannelData]: Transcripts only with video_id (excluding metadata).
        """
        
        return await self.fetcher.fetch()

    async def fetch_snippets(self) -> list[ChannelData] | None:
        """
        Returns the raw snippet data (metadata and video IDs) retrieved from the YouTube Data API.

        Returns:
            list[ChannelData] | None: An object containing video metadata and IDs.
        """

        return [
            ChannelData(
                video_id=snippet.video_id,
                transcripts=None,
                metadata=snippet
            )
            for snippet in self.snippets
        ]

    @property
    def video_ids(self) -> list[str]:
        """
        List of video IDs fetched from the YouTube channel or provided directly.

        Returns:
            list[str]: Video ID strings.
        """
        
        return self._get_video_ids()

    @property
    def metadata(self) -> list[DLSnippet] | None:
        """
        Metadata for each video, such as title, duration, and description.

        Returns:
            list[DLSnippet] | None: List of Snippet objects containing video metadata.
        """
        return [snippet for snippet in self.snippets]
    
    def _get_video_ids(self) -> list[str]:
        """
        Returns list of channel video ids.
        """
        return [snippet.video_id for snippet in self.snippets]


from abc import ABC, abstractmethod
from pathlib import Path
from ytfetcher.models.channel import ChannelData
from ytfetcher.exceptions import NoDataToExport, OutputDirectoryNotFoundError
from typing import Literal, Sequence, get_args, Any
import json
import csv
import logging
import warnings

logger = logging.getLogger(__name__)

METEDATA_LIST = Literal['title', 'description', 'url', 'duration', 'view_count', 'thumbnails']

DEFAULT_METADATA = get_args(METEDATA_LIST)

class Exporter:
    """
    DEPRECATED: Use TXTExporter, JSONExporter, or CSVExporter instead.\n
    This class is kept for backward compatibility and will be removed in a future versions.

    Parameters:
        channel_data (list[ChannelData]): The transcript and metadata to export.
        allowed_metadata_list (list): Metadata fields to include (e.g., ['title', 'description']).
        timing (bool): Whether to include start/duration timing in exports.
        filename (str): Output filename without extension.
        output_dir (str | None): Directory to export files into. Defaults to current working directory.

    Raises:
        NoDataToExport: If no data is provided.
        OutputDirectoryNotFoundError: If specified path cannot found.
    """
    def __init__(self, *args, **kwargs):
        warnings.warn(
            "The 'Exporter' class is deprecated and will be removed in future versions. Use TXTExporter, JSONExporter, or CSVExporter instead.",
            DeprecationWarning,
            stacklevel=2
        )
        self.args = args
        self.kwargs = kwargs

    def export_as_txt(self) -> None:
        return TXTExporter(*self.args, **self.kwargs).write()
    def export_as_json(self) -> None:
        return JSONExporter(*self.args, **self.kwargs).write()
    def export_as_csv(self) -> None:
        return CSVExporter(*self.args, **self.kwargs).write()

class BaseExporter(ABC):
    """
    Handles exporting YouTube transcript and metadata to various formats: TXT, JSON, and CSV.

    Supports customization of which metadata fields to include and whether to include transcript timing.

    Parameters:
        channel_data (list[ChannelData]): The transcript and metadata to export.
        allowed_metadata_list (list): Metadata fields to include (e.g., ['title', 'description']).
        timing (bool): Whether to include start/duration timing in exports.
        filename (str): Output filename without extension.
        output_dir (str | None): Directory to export files into. Defaults to current working directory.

    Raises:
        NoDataToExport: If no data is provided.
        OutputDirectoryNotFoundError: If specified path cannot found.
    """
    def __init__(self, channel_data: list[ChannelData], allowed_metadata_list: Sequence[METEDATA_LIST] = DEFAULT_METADATA, timing: bool = True, filename: str = 'data', output_dir: str | None = None):
        self.channel_data = channel_data
        self.allowed_metadata_list = allowed_metadata_list
        self.timing = timing
        self.filename = filename
        self.output_dir = Path(output_dir) if output_dir else Path.cwd()

        if not self.channel_data:
            raise NoDataToExport("No data to export.")
        
        if not self.output_dir.exists():
            raise OutputDirectoryNotFoundError("System path could not found.")

    @abstractmethod
    def write(self) -> None:
        pass

    def _initialize_output_path(self, export_type: Literal['txt', 'json', 'csv'] = 'txt') -> Path:
        self.output_dir.mkdir(parents=True, exist_ok=True)
        output_path = self.output_dir / f"{self.filename}.{export_type}"
        
        logger.info(f"Writing as {export_type} file, output path: {output_path}")

        return output_path

class TXTExporter(BaseExporter):
    """
    Exports the data as a plain text file, including transcript and metadata.
    """
    def __init__(self, channel_data, allowed_metadata_list = DEFAULT_METADATA, timing = True, filename = 'data', output_dir = None):
        super().__init__(channel_data, allowed_metadata_list, timing, filename, output_dir)
    
    def write(self):
        output_path = self._initialize_output_path(export_type='txt')
        with open(output_path, 'w', encoding='utf-8') as file:
            for data in self.channel_data:
                file.write(f"Transcript for {data.video_id}:\n")

                for metadata in self.allowed_metadata_list:
                    if data.metadata:
                        file.write(f'{metadata} --> {getattr(data.metadata, metadata)}\n')
                
                
                self._write_transcripts(file=file, data=data)

                self._write_comments(file=file, data=data)
    
    def _write_transcripts(self, file, data: ChannelData) -> None:
        if not data.transcripts: return

        for transcript in data.transcripts:
            if self.timing:
                file.write(f"{transcript.start} --> {transcript.start + transcript.duration}\n")
            file.write(f"{transcript.text}\n")
        file.write("\n")

    def _write_comments(self, file, data: ChannelData) -> None:
        if not data.comments: return

        for comment in data.comments:
            file.write(f"Comments for {data.video_id}\nComment --> {comment.text}\nAuthor --> {comment.author}\nLikes --> {comment.like_count}\nTime Text --> {comment.time_text}")
        file.write("\n")

class JSONExporter(BaseExporter):
    """
    Exports the data as a structured JSON file.
    """
    def __init__(self, channel_data, allowed_metadata_list = DEFAULT_METADATA, timing = True, filename = 'data', output_dir = None):
        super().__init__(channel_data, allowed_metadata_list, timing, filename, output_dir)
    
    def write(self):
        output_path = self._initialize_output_path(export_type='json')
        export_data = []

        with open(output_path, 'w', encoding='utf-8') as file:
            for data in self.channel_data:
                video_data = {
                    "video_id": data.video_id,
                    **{field: getattr(data.metadata, field) for field in self.allowed_metadata_list if data.metadata and hasattr(data.metadata, field)},
                }

                self._write_transcripts(data=data, video_data=video_data)

                self._write_comments(data=data, video_data=video_data)

                export_data.append(video_data)

        with open(output_path, 'w', encoding='utf-8') as file:
            json.dump(export_data, file, indent=2, ensure_ascii=False)
    
    def _write_transcripts(self, data: ChannelData, video_data: dict[str, Any]) -> None:
        if not data.transcripts: return

        video_data['transcript'] = [
            {
                **({"start": transcript.start, "duration": transcript.duration} if self.timing else {}),
                "text": transcript.text
            }
            for transcript in data.transcripts
        ]
    
    def _write_comments(self, data: ChannelData, video_data: dict[str, Any]) -> None:
        if not data.comments: return

        video_data['comments'] = [
            {
                "comment": comment.text,
                "author": comment.author,
                "time_text": comment.time_text,
                "like_count": comment.like_count
            }
            for comment in data.comments
        ]

class CSVExporter(BaseExporter):
    """
    Exports the data as a flat CSV file, row-per-transcript-entry.
    """
    def __init__(self, channel_data, allowed_metadata_list = DEFAULT_METADATA, timing = True, filename = 'data', output_dir = None):
        super().__init__(channel_data, allowed_metadata_list, timing, filename, output_dir)
    
    def write(self):
        output_path = self._initialize_output_path(export_type='csv')

        t = ['start', 'duration']
        comments = ['comment', 'comment_author', 'comment_like_count', 'comment_time_text']
        metadata = [*self.allowed_metadata_list]
        fieldnames = ['index', 'video_id', 'text']
        fieldnames += t if self.timing else []
        fieldnames += metadata if any(d.metadata for d in self.channel_data) else []
        fieldnames += comments if any(d.comments for d in self.channel_data) else []

        with open(output_path, 'w', encoding='utf-8') as file:
            writer = csv.DictWriter(file, fieldnames=fieldnames)
            writer.writeheader()

            i = 0
            for data in self.channel_data:

                base_info = {
                    'index': i,
                    'video_id': data.video_id,
                    **{field: getattr(data.metadata, field) for field in self.allowed_metadata_list if data.metadata and hasattr(data.metadata, field)},
                    **{field: getattr(data.metadata, field) for field in self.allowed_metadata_list if data.metadata and hasattr(data.metadata, field)},
                }

                self._write_comments(data=data, writer=writer, base_info=base_info)
                i += 1

                self._write_transcripts(data=data, writer=writer, base_info=base_info)
                i += 1
                
    def _write_transcripts(self, data: ChannelData, writer, base_info: dict[str, Any]) -> None:
        if not data.transcripts: return

        for transcript in data.transcripts:
            row = {
                **base_info,
                **({"start": transcript.start, "duration": transcript.duration} if self.timing else {}),
                'text': transcript.text
            }

            writer.writerow(row)
    
    def _write_comments(self, data: ChannelData, writer, base_info: dict[str, Any]) -> None:
        if not data.comments: return
        
        for comment in data.comments:
            row = {
                **base_info,
                'comment': comment.text,
                'comment_author': comment.author,
                'comment_like_count': comment.like_count,
                'comment_time_text': comment.time_text
            }

            if data.transcripts:
                for transcript in data.transcripts:
                    row.update({
                        **({"start": transcript.start, "duration": transcript.duration} if self.timing else {}),
                        'text': transcript.text,
                    })  

            writer.writerow(row)


import argparse
import asyncio
import ast
import sys
from typing import Union
from ytfetcher._core import YTFetcher
from ytfetcher.services.exports import TXTExporter, CSVExporter, JSONExporter, BaseExporter, DEFAULT_METADATA
from ytfetcher.config.http_config import HTTPConfig
from ytfetcher.config import GenericProxyConfig, WebshareProxyConfig
from ytfetcher.models import ChannelData
from ytfetcher.utils.log import log

from argparse import ArgumentParser

class YTFetcherCLI:
    """
    YTFetcherCLI
    A command-line interface for fetching and exporting YouTube transcripts.
    This class handles the orchestration of transcript fetching operations from various YouTube sources
    (channels, videos, or playlists) and manages the export of fetched data in multiple formats.
    """
    def __init__(self, args: argparse.Namespace):
        self.args = args
    
    def _initialize_proxy_config(self) -> Union[WebshareProxyConfig, GenericProxyConfig, None]:
        proxy_config = None

        if self.args.http_proxy != "" or self.args.https_proxy != "":
            proxy_config = GenericProxyConfig(
                http_url=self.args.http_proxy,
                https_url=self.args.https_proxy,
            )

        if (
            self.args.webshare_proxy_username is not None
            or self.args.webshare_proxy_password is not None
        ):
            proxy_config = WebshareProxyConfig(
                proxy_username=self.args.webshare_proxy_username,
                proxy_password=self.args.webshare_proxy_password,
        )
            
        return proxy_config

    def _initialize_http_config(self) -> HTTPConfig:
        if self.args.http_timeout or self.args.http_headers:
            http_config = HTTPConfig(timeout=self.args.http_timeout, headers=self.args.http_headers)
            return http_config

        return HTTPConfig()
    
    async def _run_fetcher(self, factory_method: type[YTFetcher], **kwargs) -> None:
        fetcher = factory_method(
            http_config=self._initialize_http_config(),
            proxy_config=self._initialize_proxy_config(),
            **kwargs
        )

        async def get_data(comments_arg: int, comments_only_arg: int) -> list[ChannelData]:
            """
            Decides correct method and returns data based on `comments` argument.
            
            :param comments: Whether comments arg is None or not.
            :type comments: bool
            """
            if comments_arg > 0:
                return await fetcher.fetch_with_comments(max_comments=self.args.comments)

            elif comments_only_arg > 0:
                return await fetcher.fetch_comments(max_comments=self.args.comments_only)

            return await fetcher.fetch_youtube_data()

        data = await get_data(comments_arg=self.args.comments, comments_only_arg=self.args.comments_only)
        log('Fetched all transcripts.', level='DONE')
        if self.args.print:
            print(data)
        self._export(data)
        log(f"Data exported successfully as {self.args.format}", level='DONE')
    
    @staticmethod
    def _get_exporter(format_type: str) -> type[BaseExporter]:
        """
        Factory to return the correct Exporter class based on string.
        """
        registry: dict[str, type[BaseExporter]] = {
            "txt": TXTExporter,
            "json": JSONExporter,
            'csv': CSVExporter 
        }

        exporter_class = registry.get(format_type.lower())
        if not exporter_class:
            raise ValueError(f'Unsupported format {format_type}')
        
        return exporter_class

    def _export(self, channel_data: list[ChannelData]) -> None:
        exporter_class = self._get_exporter(self.args.format)
        exporter = exporter_class(
            channel_data=channel_data,
            output_dir=self.args.output_dir,
            filename=self.args.filename,
            allowed_metadata_list=self.args.metadata,
            timing=not self.args.no_timing
        )

        exporter.write()
    
    async def run(self):
        match self.args.command:
            case 'from_channel':
                log(f'Starting to fetch from channel: {self.args.channel_handle}')
                await self._run_fetcher(
                    YTFetcher.from_channel,
                    channel_handle=self.args.channel_handle,
                    max_results=self.args.max_results,
                    languages=self.args.languages,
                    manually_created=self.args.manually_created,
                )
            
            case 'from_video_ids':
                log(f'Starting to fetch from video ids: {self.args.video_ids}')
                await self._run_fetcher(
                    YTFetcher.from_video_ids,
                    video_ids=self.args.video_ids,
                    languages=self.args.languages,
                    manually_created=self.args.manually_created
                )
            
            case 'from_playlist_id':
                log(f"Starting to fetch from playlist id: {self.args.playlist_id}")
                await self._run_fetcher(
                    YTFetcher.from_playlist_id,
                    playlist_id=self.args.playlist_id,
                    languages=self.args.languages,
                    manually_created=self.args.manually_created
                )

            case _:
                raise ValueError(f"Unknown method: {self.args.command}")

def create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description="Fetch YouTube transcripts for a channel")

    subparsers = parser.add_subparsers(dest="command")

    # From Channel parsers
    parser_channel = subparsers.add_parser("from_channel", help="Fetch data from channel handle with max_results.")

    parser_channel.add_argument("-c", "--channel_handle", help="YouTube channel handle")
    parser_channel.add_argument("-m", "--max-results", type=int, default=5, help="Maximum videos to fetch")
    _create_common_arguments(parser_channel)

    # From Video Ids parsers
    parser_video_ids = subparsers.add_parser("from_video_ids", help="Fetch data from your custom video ids.")

    parser_video_ids.add_argument("-v", "--video-ids", nargs="+", help='Video id list to fetch')
    _create_common_arguments(parser_video_ids)

    # From playlist_id parsers
    parser_playlist_id = subparsers.add_parser("from_playlist_id", help="Fetch data from a specific playlist id.")

    parser_playlist_id.add_argument("-p", "--playlist-id", type=str, help='Playlist id to be fetch from.')
    _create_common_arguments(parser_playlist_id)

    return parser

def parse_args(argv=None):
    parser = create_parser()
    return parser.parse_args(args=argv)

def _create_common_arguments(parser: ArgumentParser) -> None:
    """
    Creates common arguments for parsers.
    """
    parser.add_argument("-o", "--output-dir", default=".", help="Output directory for data")
    parser.add_argument("-f", "--format", choices=["txt", "json", "csv"], default="txt", help="Export format")
    parser.add_argument("--metadata", nargs="+", default=DEFAULT_METADATA, choices=DEFAULT_METADATA, help="Allowed metadata")
    parser.add_argument("--no-timing", action="store_true", help="Do not write transcript timings like 'start', 'duration'")
    parser.add_argument("--languages", nargs="+", default=["en"], help="List of language codes in priority order (e.g. en de fr). Defaults to ['en'].")
    parser.add_argument("--manually-created", action="store_true", help="Fetch only videos that has manually created transcripts.")
    parser.add_argument("--comments", default=0, type=int, help="Add top comments to the metadata alongside with transcripts.")
    parser.add_argument("--comments-only", default=0, type=int, help="Fetch only comments with metadata.")
    parser.add_argument("--print", action="store_true", help="Print data to console.")
    parser.add_argument("--filename", default="data", help="Decide filename to be exported.")
    parser.add_argument("--http-timeout", type=float, default=4.0, help="HTTP timeout for requests.")
    parser.add_argument("--http-headers", type=ast.literal_eval, help="Custom http headers.")
    parser.add_argument("--webshare-proxy-username", default=None, type=str, help='Specify your Webshare "Proxy Username" found at https://dashboard.webshare.io/proxy/settings')
    parser.add_argument("--webshare-proxy-password", default=None, type=str, help='Specify your Webshare "Proxy Password" found at https://dashboard.webshare.io/proxy/settings')
    parser.add_argument("--http-proxy", default="", metavar="URL", help="Use the specified HTTP proxy.")
    parser.add_argument("--https-proxy", default="", metavar="URL", help="Use the specified HTTPS proxy.")

def main():
    args = parse_args(sys.argv[1:])
    cli = YTFetcherCLI(args=args)
    asyncio.run(cli.run())

if __name__ == "__main__":
    main()



import logging

def enable_default_config(level=logging.INFO):
    """
    Simple method for enabling basic logging.
    Args:
        level: Log level. Default to INFO
    """
    logging.basicConfig(
        level=level,
        format="%(asctime)s [%(levelname)s] %(name)s: %(message)s",
    )


import yt_dlp
import concurrent.futures
import logging
from ytfetcher.models.channel import DLSnippet, Comment
from tqdm import tqdm
from abc import ABC, abstractmethod
from urllib.parse import urlparse, parse_qs
from ytfetcher.utils.log import log
from typing import Any, cast

logger = logging.getLogger(__name__)

class CommentFetcher:
    def __init__(self, max_comments: int = 20, max_workers: int = 20):
        self.max_comments = max_comments
        self.max_workers = max_workers
    
    def fetch(self, video_ids: list[str]) -> list[list[Comment]]:
        log(f"Starting to fetch comments for {len(video_ids)} videos...", level='INFO')
        
        with concurrent.futures.ThreadPoolExecutor(max_workers=self.max_workers) as executor:
            futures = [executor.submit(self._fetch_single, video_id) for video_id in video_ids]

            results = []
            for future in tqdm(concurrent.futures.as_completed(futures), total=len(video_ids), desc="Fetching Comments"):
                results.append(future.result())
                
            return results
            
    def _fetch_single(self, video_id: str) -> list[Comment]:
        video_url = f'https://www.youtube.com/watch?v={video_id}'
        
        ydl_opts_deep = {
            "quiet": True,
            "skip_download": True,
            "force_empty": True,
            "getcomments": True,
            "no_warnings": True,
            "extractor_args": {
                "youtube": {
                    "max_comments": [
                        str(self.max_comments),
                        "all",
                        "0",
                        "0"
                    ],
                    "comment_sort": "top"
                }
            }
        }
                
        with yt_dlp.YoutubeDL(ydl_opts_deep) as ydl: #type: ignore[arg-type]
            info_dict = ydl.extract_info(video_url, download=False)
            data = cast(list[dict[str, Any]], info_dict.get('comments', None) or [])
            return [Comment.model_validate(comment) for comment in data]

class BaseYoutubeDLFetcher(ABC):
    """
    Abstract base class for YouTube data fetching using yt_dlp.

    Provides common setup utilities for subclasses that fetch
    metadata from channels, playlists, or specific videos.

    Attributes:
        max_results (int): The maximum number of results to fetch.
    """

    def __init__(self, max_results: int = 50):
        self.max_results = max_results

    @abstractmethod
    def fetch(self) -> list[DLSnippet]:
        """
        Abstract method to be implemented by subclasses.

        Should contain the logic to fetch data from YouTube.
        """
        pass

    def _setup_ydl_opts(self, **extra_opts) -> dict:
        """
        Prepare yt_dlp options with safe defaults for metadata extraction.

        Args:
            **extra_opts: Additional yt_dlp options to override or extend defaults.

        Returns:
            dict: The complete yt_dlp options dictionary.
        """
        base_opts = {
            "quiet": True,
            "skip_download": True,
            "extract_flat": True,
            "no_warnings": True,
        }
        base_opts.update(extra_opts)
        return base_opts

    def _to_snippets(self, entries: list[dict[str, Any]]) -> list[DLSnippet]:
        """
        Convert yt_dlp raw entries into DLSnippet objects.

        Args:
            entries (list[dict]): List of yt_dlp video entries.

        Returns:
            list[DLSnippet]: List of structured DLSnippet objects.
        """
        return [DLSnippet.model_validate(entry) for entry in entries]


class ChannelFetcher(BaseYoutubeDLFetcher):
    """
    Fetches recent videos from a YouTube channel.

    Supports both direct channel handles (e.g. "@caseoh_")
    and full channel URLs.
    Args:
        channel_handle (str): The channel handle or URL.
        max_results (int): Maximum number of videos to fetch.
    """

    def __init__(self, channel_handle: str, max_results: int = 50):
        super().__init__(max_results)
        self.channel_handle = channel_handle

        if "https://" in channel_handle:
            self.channel_handle = self._find_channel_handle_from_url(channel_handle)

    def fetch(self) -> list[DLSnippet]:
        ydl_opts = self._setup_ydl_opts(playlistend=self.max_results)
        url = f"https://www.youtube.com/@{self.channel_handle.replace('@', '').strip()}/videos"

        with yt_dlp.YoutubeDL(ydl_opts) as ydl: #type: ignore[arg-type]
            info = ydl.extract_info(url, download=False)
            entries = cast(list[dict[str, Any]], info.get("entries", []))
            return self._to_snippets(entries)

    @staticmethod
    def _find_channel_handle_from_url(url: str) -> str:
        """
        Extract the channel handle from a full YouTube channel URL.

        Handles URLs such as:
            - https://www.youtube.com/@handle
            - https://www.youtube.com/@handle/
            - https://www.youtube.com/@handle/featured
            - https://www.youtube.com/@handle/videos

        Args:
            url (str): Full YouTube channel URL.

        Returns:
            str: The extracted channel handle (e.g. "handle").

        Raises:
            ValueError: If no valid handle can be extracted.
        """
        log("Got full URL, trying to extract channel handle. If it fails, try providing only the handle.", level="WARNING")

        parsed = urlparse(url)
        path = parsed.path

        if "@" in path:
            handle = path.split("@", 1)[1].split("/")[0]
            return handle.strip()

        raise ValueError(f"Could not extract channel handle from URL: {url}")


class PlaylistFetcher(BaseYoutubeDLFetcher):
    """
    Fetches all videos from a given YouTube playlist.

    Supports both playlist IDs and full playlist URLs.
    Args:
        playlist_id (str): The playlist ID or full URL.
        max_results (int): Maximum number of videos to fetch.
    """

    def __init__(self, playlist_id: str, max_results: int = 50):
        super().__init__(max_results)
        self.playlist_id = playlist_id

        if "https://" in playlist_id:
            self.playlist_id = self._find_playlist_id_from_url(url=playlist_id)

    def fetch(self) -> list[DLSnippet]:
        ydl_opts = self._setup_ydl_opts(playlistend=self.max_results)
        url = f"https://www.youtube.com/playlist?list={self.playlist_id.strip()}"

        with yt_dlp.YoutubeDL(ydl_opts) as ydl: #type: ignore[arg-type]
            info = ydl.extract_info(url, download=False)
            entries = cast(list[dict[str, Any]], info.get("entries", []))
            return self._to_snippets(entries)

    @staticmethod
    def _find_playlist_id_from_url(url: str) -> str:
        """
        Extract the playlist ID from a YouTube playlist URL.

        Handles URLs such as:
            - https://www.youtube.com/playlist?list=PL12345
            - https://www.youtube.com/playlist?list=PL12345&si=abcd
            - https://youtube.com/watch?v=abc123&list=PL12345

        Args:
            url (str): Full YouTube playlist URL.

        Returns:
            str: The extracted playlist ID (e.g. "PL12345").

        Raises:
            ValueError: If no valid playlist ID can be found.
        """
        log("Got full URL, trying to extract playlist ID. If it fails, try providing only playlist ID.", level="WARNING")

        parsed = urlparse(url)
        query_params = parse_qs(parsed.query)
        playlist_id_list = query_params.get("list")

        if playlist_id_list and len(playlist_id_list) > 0:
            return playlist_id_list[0].strip()

        raise ValueError(f"Could not extract playlist ID from URL: {url}")


class VideoListFetcher(BaseYoutubeDLFetcher):
    """
    Fetches metadata for one or more specific YouTube videos.
    Args:
        video_ids (list[str]): List of YouTube video IDs.
    """

    def __init__(self, video_ids: list[str]):
        self.video_ids = video_ids

    def fetch(self) -> list[DLSnippet]:
        ydl_opts = self._setup_ydl_opts()
        results: list[dict[str, Any]] = []

        with yt_dlp.YoutubeDL(ydl_opts) as ydl: #type: ignore[arg-type]
            for video_id in tqdm(self.video_ids, desc="Extracting metadata", unit="video"):
                url = f"https://www.youtube.com/watch?v={video_id}"
                info = cast(dict[str, Any], ydl.extract_info(url, download=False))
                if info:
                    results.append(info)

        return self._to_snippets(entries=results)


def get_fetcher(
    channel_handle: str | None = None,
    playlist_id: str | None = None,
    video_ids: list[str] | None = None,
    max_results: int = 50,
) -> BaseYoutubeDLFetcher:
    """
    Factory function that returns the correct fetcher
    based on provided parameters.

    Args:
        channel_handle (str | None): YouTube channel handle or URL.
        playlist_id (str | None): YouTube playlist ID or URL.
        video_ids (list[str] | None): List of specific video IDs.
        max_results (int): Maximum number of videos to fetch.

    Returns:
        BaseYoutubeDLFetcher: An appropriate fetcher subclass instance.

    Raises:
        ValueError: If no valid input was provided.
    """
    if playlist_id:
        return PlaylistFetcher(playlist_id, max_results)
    elif channel_handle:
        return ChannelFetcher(channel_handle, max_results)
    elif video_ids:
        return VideoListFetcher(video_ids)
    raise ValueError("No YoutubeDLFetcher found.")



from pydantic import BaseModel, Field, model_validator, ConfigDict

class Comment(BaseModel):

    model_config = ConfigDict(populate_by_name=True)

    id: str
    text: str
    like_count: int | None = None
    author: str | None = None
    time_text: str | None = Field(alias='_time_text', default=None)

class DLSnippet(BaseModel):

    model_config = ConfigDict(populate_by_name=True)

    video_id: str = Field(alias='id')
    title: str
    description: str | None = None
    url: str | None = None
    duration: float | None = None
    view_count: int | None = None
    thumbnails: list[dict] | None = None

    @model_validator(mode='after')
    def validate_url(self) -> 'DLSnippet':
        """If URL is missing, build it using the video_id."""
        if not self.url:
            self.url = f"https://youtube.com/watch?v={self.video_id}"
        return self

class Transcript(BaseModel):
    text: str
    start: float
    duration: float
    
class VideoTranscript(BaseModel):
    video_id: str
    transcripts: list[Transcript]

    def to_dict(self) -> dict:
        return self.model_dump()

class ChannelData(BaseModel):
    video_id: str
    transcripts: list[Transcript] | None = None
    metadata: DLSnippet | None = None
    comments: list[Comment] | None = None

    def to_dict(self) -> dict:
        return self.model_dump(exclude_none=True)


from youtube_transcript_api._errors import NoTranscriptFound, VideoUnavailable, TranscriptsDisabled
from youtube_transcript_api import YouTubeTranscriptApi
from youtube_transcript_api.proxies import ProxyConfig
from ytfetcher.models.channel import ChannelData, VideoTranscript, Transcript
from ytfetcher.config.http_config import HTTPConfig
from ytfetcher.utils.log import log
from concurrent.futures import ThreadPoolExecutor
from tqdm import tqdm  # type: ignore
from typing import Iterable
import asyncio
import requests
import logging
import re

logger = logging.getLogger(__name__)
class TranscriptFetcher:
    """
    Asynchronously fetches transcripts for a list of YouTube video IDs 
    using the YouTube Transcript API.

    Transcripts are fetched concurrently using threads, while optionally 
    supporting proxy configurations and custom HTTP settings.

    Args:
        video_ids (list[str]): 
            List of YouTube video IDs to fetch transcripts for.

        http_config (HTTPConfig): 
            Optional HTTP configuration (e.g., headers, timeout).

        proxy_config (ProxyConfig | None): 
            Optional proxy configuration for the YouTube Transcript API.

        languages (Iterable[str]): 
            A list of language codes in descending priority. 
            For example, if this is set to ["de", "en"], it will first try 
            to fetch the German transcript ("de") and then the English one 
            ("en") if it fails. Defaults to ["en"].
    """

    def __init__(self, video_ids: list[str], http_config: HTTPConfig = HTTPConfig(), proxy_config: ProxyConfig | None = None, languages: Iterable[str] = ("en",), manually_created: bool = False):
        self.video_ids = video_ids
        self.languages = languages
        self.manually_created = manually_created
        self.executor = ThreadPoolExecutor(max_workers=30)
        self.proxy_config = proxy_config
        self.http_client = requests.Session()

        # Initialize client
        self.http_client.headers = http_config.headers

    async def fetch(self) -> list[ChannelData]:
        """
        Asynchronously fetches transcripts for all provided video IDs.

        Transcripts are fetched using threads wrapped in asyncio. Results are streamed as they are completed,
        and errors like `NoTranscriptFound`, `TranscriptsDisabled`, or `VideoUnavailable` are silently handled.

        Returns:
            list[VideoTranscript]: A list of successful transcripts from list of videos with video_id information.
        """

        async def run_in_thread(video_id: str):
            return await asyncio.to_thread(self._fetch_single, video_id)

        tasks = [run_in_thread(video_id) for video_id in self.video_ids]

        channel_data = await self._build_channel_data(tasks)
        
        if not channel_data and self.manually_created: log("No manually created transcripts found!", level="ERROR")

        return channel_data

    def _fetch_single(self, video_id: str) -> VideoTranscript | None:
        """
        Fetches a single transcript and returns structured data.

        Handles known errors from the YouTube Transcript API gracefully.
        Logs warnings for unavailable or disabled transcripts.

        Parameters:
            video_id (str): The ID of the YouTube video to fetch.

        Returns:
            VideoTranscript | None: A dictionary with transcript and video_id,
                         or None if transcript is unavailable.
        """
        try:
            yt_api = YouTubeTranscriptApi(http_client=self.http_client, proxy_config=self.proxy_config)
            transcript: list[Transcript] | None = self._decide_fetch_method(yt_api, video_id)

            if not transcript: return None

            cleaned_transcript = self._clean_transcripts(transcript)
            logger.info(f'{video_id} fetched.')
            return VideoTranscript(
                video_id=video_id,
                transcripts=cleaned_transcript
            )
        except (NoTranscriptFound, VideoUnavailable, TranscriptsDisabled) as e:
            logger.warning(e)
            return None
        except Exception as e:
            logger.warning(f'Error while fetching transcript from video: {video_id} ', e)
            return None
    
    def _decide_fetch_method(self, yt_api: YouTubeTranscriptApi, video_id: str) -> list[Transcript] | None:
        """
        Decides correct fetch method based on manually created flag.
        Args:
            yt_api(YouTubeTranscriptApi): Ytt api instance.
            video_id(str): Video id for current video.
        Returns:
            Optional[list[Transcript]]:
                A list of transcript entries as dictionaries if available, 
                otherwise `None` when no transcript is found.
        """
        if self.manually_created:
            try:
                raw_transcripts = yt_api.list(video_id).find_manually_created_transcript(language_codes=self.languages).fetch().to_raw_data()
                return self._convert_to_transcript_object(raw_transcripts)
            except NoTranscriptFound:
                logger.info(f"Couldn't found manually created transcript for {video_id}")
                return None
        
        raw_transcripts = yt_api.fetch(video_id, languages=self.languages).to_raw_data()
        return self._convert_to_transcript_object(raw_transcripts)
    
    @staticmethod
    async def _build_channel_data(tasks: list) -> list[ChannelData]:
        """
        Builds list of `ChannelData` from all tasks from completed thread with progress support.
        Args:
            tasks: List of completed tasks.
        Returns:
            list[ChannelData]: ChannelData list contains transcripts.
        """
        channel_data: list[ChannelData] = []

        for coro in tqdm(asyncio.as_completed(tasks), total=len(tasks), desc="Fetching transcripts", unit='transcript'):
            result: VideoTranscript = await coro
            if result:
                channel_data.append(
                    ChannelData(
                        video_id=result.video_id,
                        transcripts=result.transcripts,
                        metadata=None
                    )
                )
        return channel_data

    @staticmethod
    def _clean_transcripts(transcripts: list[Transcript]) -> list[Transcript]:
        """
        Cleans unnecessary text from transcripts like [Music], [Applause], etc.
        Returns:
            list[Transcript]: list of Transcript objects.
        """
        for entry in transcripts:

            # Remove unnecessary text patterns like [Music], [Applause], etc.
            cleaned_text = re.sub(r'\[.*?\]', '', entry.text)

            # Remove leading '>>' markers (and optional spaces)
            cleaned_text = re.sub(r'^\s*>>\s*', '', cleaned_text)

            # Remove extra whitespace
            cleaned_text = ' '.join(cleaned_text.split())

            # Update the transcript text
            entry.text = cleaned_text

        return transcripts
    
    @staticmethod
    def _convert_to_transcript_object(transcript_dict: list[dict]) -> list[Transcript]:
        return [
            Transcript(
            text=transcript['text'],
            start=transcript['start'],
            duration=transcript['duration']
            )
        for transcript in transcript_dict
        ]




from typing import Literal
import colorama

LEVEL = Literal['INFO', 'WARNING', 'DONE', 'ERROR']

def decide_color(level: LEVEL):
    """
    Chooses color based on message level.
    Args:
        level(LEVEL): Message level.
    """
    match level:
        case 'INFO':
            return colorama.Fore.CYAN
        case 'WARNING':
            return colorama.Fore.YELLOW
        case 'ERROR':
            return colorama.Fore.RED
        case 'DONE':
            return colorama.Fore.GREEN

def log(message: str, level: LEVEL = ('INFO')):
    """
    Print message with different levels and colors. ex: (INFO, WARNING, DONE, ERROR).
    Args:
        message(str): The message to be printed.
        level(LEVEL): Priority level. Default to INFO
    """
    decided_color = decide_color(level=level)
    return print(decided_color + f'[{level}] ' + colorama.Style.RESET_ALL + message)